@@ -2,13 +2,14 @@
 
 import os
 import sys
+import traceback
 from pprint import pprint
 from argparse import ArgumentParser
 import numpy as np
 import onnx
 import onnx_graphsurgeon as gs
 from onnx_graphsurgeon.ir.tensor import Constant
-from typing import Tuple, Optional
+from typing import Tuple, Optional, List
 
 class Color:
     BLACK          = '\033[30m'
@@ -37,20 +38,29 @@ class Color:
 
 
 def shrinking(
-    input_onnx_file_path: str,
-    output_onnx_file_path: str,
+    input_onnx_file_path: Optional[str] = '',
+    output_onnx_file_path: Optional[str] = '',
+    onnx_graph: Optional[onnx.ModelProto] = None,
     mode: Optional[str] = 'shrink',
+    forced_extraction_op_names: List[str] = [],
     non_verbose: Optional[bool] = False,
-) -> Tuple[onnx.ModelProto, str]:
+) -> Tuple[onnx.ModelProto, List[str]]:
 
     """
     Parameters
     ----------
-    input_onnx_file_path: str
-        Input onnx file path.
+    input_onnx_file_path: Optional[str]
+        Input onnx file path.\n\
+        Either input_onnx_file_path or onnx_graph must be specified.
 
-    output_onnx_file_path: str
-        Outpu onnx file path.
+    output_onnx_file_path: Optional[str]
+        Outpu onnx file path.\n\
+        If output_onnx_file_path is not specified, no .onnx file is output.
+
+    onnx_graph: Optional[onnx.ModelProto]
+        onnx.ModelProto.\n\
+        Either input_onnx_file_path or onnx_graph must be specified.\n\
+        onnx_graph If specified, ignore input_onnx_file_path and process onnx_graph.
 
     mode: Optional[str]
         Constant Value Compression Mode.\n\
@@ -61,20 +71,37 @@ def shrinking(
             Instead of the smallest model body size, the file loading overhead is greater.\n\
         Default: shrink
 
+    forced_extraction_op_names: List[str]
+        Extracts the constant value of the specified OP name to .npy regardless of the mode specified.\n\
+        e.g. ['aaa','bbb','ccc']
+
     non_verbose: Optional[bool]
-        Do not show all information logs. Only error logs are displayed.
+        Do not show all information logs. Only error logs are displayed.\n\
+        Default: False
 
     Returns
     -------
     shrunken_graph: onnx.ModelProto
         Shrunken onnx ModelProto
 
-    npy_file_paths:
+    npy_file_paths: List[str]
         List of paths to externally output .npy files.
         An empty list is always returned when in 'shrink' mode.
     """
+
+    if not input_onnx_file_path and not onnx_graph:
+        print(
+            f'{Color.RED}ERROR:{Color.RESET} '+
+            f'One of input_onnx_file_path or onnx_graph must be specified.'
+        )
+        sys.exit(1)
+
     # Loading Graphs
-    graph = gs.import_onnx(onnx.load(input_onnx_file_path))
+    # onnx_graph If specified, onnx_graph is processed first
+    if not onnx_graph:
+        graph = gs.import_onnx(onnx.load(input_onnx_file_path))
+    else:
+        graph = gs.import_onnx(onnx_graph)
 
     # Constant Value Extraction
     constants = {}
@@ -86,6 +113,22 @@ def shrinking(
                 continue
             if np.isscalar(graph_node_input.values):
                 continue
+
+            # Try downcast
+            ### INT64 -> INT32
+            if graph_node_input.values.dtype == np.int64:
+                orig = graph_node_input.values
+                dist = graph_node_input.values.astype(np.int32)
+                if (orig == dist).all():
+                    graph_node_input.values = dist
+
+            ### Float64 -> Float32
+            if graph_node_input.values.dtype == np.float64:
+                orig = graph_node_input.values
+                dist = graph_node_input.values.astype(np.float32)
+                if (orig == dist).all():
+                    graph_node_input.values = dist
+
             constants[graph_node_input.name] = graph_node_input
     if not non_verbose:
         print(
@@ -95,13 +138,13 @@ def shrinking(
 
     constants_list = list(constants.items())
 
+    # Aggregate list generation
+    aggregate_constants_name_list = {}
+    aggregate_constants_value_list = {}
+
     # Processing is performed only when the number of constant tensors to be processed is 2 or more
     if len(constants_list) >= 2:
-        # Aggregate list generation
-        aggregate_constants_name_list = {}
-        aggregate_constants_value_list = {}
         aggregate_constants_matched_list = [False] * len(constants_list)
-
         for comparator_idx in range(0, len(constants_list)-1):
             if not aggregate_constants_matched_list[comparator_idx]:
                 for comparison_dest_idx in range(comparator_idx+1, len(constants_list)):
@@ -121,12 +164,20 @@ def shrinking(
     aggregate_constants
     {'425': ['434', '5506'], '1646': ['1910', '2608', '2872', '3570', '3834'], '5550': ['4107']}
     """
+    # Automatic aggregation of constants
     npy_file_paths = []
     for layer_name_quoted_src, layer_name_quoted_dists in aggregate_constants_name_list.items():
         i = None
         if mode == 'npy':
             # Export constant values to a numpy file
-            external_file_name = f"{os.path.splitext(os.path.basename(output_onnx_file_path))[0]}_exported_{layer_name_quoted_src.replace(':','_').replace(';','_').replace('/','_').replace(',','_')}.npy"
+            external_file_name = ''
+            if output_onnx_file_path:
+                external_file_name = \
+                    f"{os.path.splitext(os.path.basename(output_onnx_file_path))[0]}" + \
+                    f"_exported_{layer_name_quoted_src.replace(':','_').replace(';','_').replace('/','_').replace(',','_')}.npy"
+            else:
+                external_file_name = \
+                    f"exported_{layer_name_quoted_src.replace(':','_').replace(';','_').replace('/','_').replace(',','_')}.npy"
             np.save(
                 external_file_name,
                 aggregate_constants_value_list[layer_name_quoted_src].values
@@ -150,6 +201,61 @@ def shrinking(
                         graph.nodes[graph_node_idx].inputs[input_idx] = i
 
     graph.cleanup().toposort()
+
+
+    # Searches the entire model by the OP name specified in forced_extraction_op_names,
+    # and if an OP with a matching name is found, the constant is forced to be extracted to the .npy file.
+
+    # Constant Value Extraction
+    constants = {}
+    graph_nodes = [node for node in graph.nodes if node.name in forced_extraction_op_names]
+    for graph_node in graph_nodes:
+        for graph_node_input in graph_node.inputs:
+            if not isinstance(graph_node_input, Constant):
+                continue
+            if len(graph_node_input.shape) == 0:
+                continue
+            if np.isscalar(graph_node_input.values):
+                continue
+            constants[graph_node_input.name] = graph_node_input
+    if not non_verbose:
+        print(
+            f'{Color.GREEN}INFO:{Color.RESET} '+
+            f'Forced-extraction number of constant values to be studied: {len(constants)}'
+        )
+
+    constants_list = list(constants.items())
+
+    for constant_key, constant_value in constants_list:
+        i = None
+        # Export constant values to a numpy file
+        external_file_name = ''
+        if output_onnx_file_path:
+            external_file_name = \
+                f"{os.path.splitext(os.path.basename(output_onnx_file_path))[0]}" + \
+                f"_exported_{constant_key.replace(':','_').replace(';','_').replace('/','_').replace(',','_')}.npy"
+        else:
+            external_file_name = \
+                f"exported_{constant_key.replace(':','_').replace(';','_').replace('/','_').replace(',','_')}.npy"
+        np.save(
+            external_file_name,
+            constant_value.values
+        )
+        npy_file_paths.append(external_file_name)
+        # Generate Inputs
+        i = gs.Variable(
+            name=external_file_name,
+            dtype=constant_value.values.dtype,
+            shape=constant_value.values.shape,
+        )
+        for graph_node_idx, graph_node in enumerate(graph.nodes):
+            for input_idx, graph_node_input in enumerate(graph_node.inputs):
+                if graph_node_input.name == constant_key:
+                    graph.nodes[graph_node_idx].inputs[input_idx] = i
+        graph.inputs.append(i)
+
+    graph.cleanup().toposort()
+
     if not non_verbose:
         print(f'{Color.GREEN}INFO:{Color.RESET} Results:')
         pprint(aggregate_constants_name_list)
@@ -162,7 +268,7 @@ def shrinking(
     new_model = None
     try:
         new_model = onnx.shape_inference.infer_shapes(shrunken_graph)
-    except:
+    except Exception as e:
         new_model = shrunken_graph
         if not non_verbose:
             print(
@@ -170,9 +276,12 @@ def shrinking(
                 'The input shape of the next OP does not match the output shape. '+
                 'Be sure to open the .onnx file to verify the certainty of the geometry.'
             )
+            tracetxt = traceback.format_exc().splitlines()[-1]
+            print(f'{Color.YELLOW}WARNING:{Color.RESET} {tracetxt}')
 
     # Save
-    onnx.save(new_model, f'{output_onnx_file_path}')
+    if output_onnx_file_path:
+        onnx.save(new_model, f'{output_onnx_file_path}')
 
     return new_model, npy_file_paths
 
@@ -206,6 +315,16 @@ def main():
             Instead of the smallest model body size, the file loading overhead is greater. \
             Default: shrink"
     )
+    parser.add_argument(
+        '--forced_extraction_op_names',
+        type=str,
+        default='',
+        help="\
+            Extracts the constant value of the specified OP name to .npy \
+            regardless of the mode specified. \
+            Specify the name of the OP, separated by commas. \
+            e.g. --forced_extraction_op_names aaa,bbb,ccc"
+    )
     parser.add_argument(
         '--non_verbose',
         action='store_true',
@@ -224,8 +343,19 @@ def main():
         )
         sys.exit(1)
 
+    forced_extraction_op_names = args.forced_extraction_op_names.strip(' ,').replace(' ','').split(',')
+
     # Model shrink
-    shrunken_graph, npy_file_paths = shrinking(args.input_onnx_file_path, args.output_onnx_file_path, args.mode, args.non_verbose)
+    shrunken_graph, npy_file_paths = shrinking(
+        input_onnx_file_path=args.input_onnx_file_path,
+        output_onnx_file_path=args.output_onnx_file_path,
+        mode=args.mode,
+        forced_extraction_op_names=forced_extraction_op_names,
+        non_verbose=args.non_verbose
+    )
 
     if not args.non_verbose:
         print(f'{Color.GREEN}INFO:{Color.RESET} Finish!')
+
+if __name__ == '__main__':
+    main()
\ No newline at end of file
